/* "tempomap"
 *  -- Uses a 'click-track' added to a midifile to correct the tempo
 *  and align notes within bar divisions
 *
 * Original version by Div Slobin June 2006
 * Revision by Pete Goodeve February 2010
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <midifile.h>

/*#define DPRINTF(r) printf r*/
#define DPRINTF(r)


static char * usage_text
 = "Usage:  %s --clicktrack|-c <n> [options] <filename.mid>\n"
  "Options (long|short):\n"
  "   --out|-o <filename.mid>\n"
  "   --beats-per-click|-b <n>\n"
  "   --beats-in-bar|-n <n>\n"
  "   --replace|-r\n"
  " Tracks are counted from 0\n Default: beats/click=1, beats/bar=4\n"
  " If output file is omitted and '--replace' is set, original will be overwritten.\n"
  " A 'lead-in' bar (or bars) will be included for notes before the first click\n" ;
 
static void usage(char *program_name)
{
	fprintf(stderr, usage_text, program_name);
	exit(1);
}

static MidiFileEvent_t get_first_click_event(MidiFile_t midi_file, int click_track_number)
{
	MidiFileTrack_t click_track = MidiFile_getTrackByNumber(midi_file, click_track_number, 0);
	MidiFileEvent_t click_event = MidiFileTrack_getFirstEvent(click_track);

	while ((click_event != NULL) && (! MidiFileEvent_isNoteStartEvent(click_event)))
	{
		click_event = MidiFileEvent_getNextEventInTrack(click_event);
	}

	return click_event;
}

static MidiFileEvent_t get_next_click_event(MidiFileEvent_t click_event)
{
	do
	{
		click_event = MidiFileEvent_getNextEventInTrack(click_event);
	}
	while ((click_event != NULL) && (! MidiFileEvent_isNoteStartEvent(click_event)));

	return click_event;
}

static MidiFileEvent_t clone_event(MidiFileEvent_t input_event, MidiFile_t output_midi_file, long output_event_tick)
{
	switch (MidiFileEvent_getType(input_event))
	{
		case MIDI_FILE_EVENT_TYPE_NOTE_OFF:
		{
			return MidiFileTrack_createNoteOffEvent(MidiFile_getTrackByNumber(output_midi_file, MidiFileTrack_getNumber(MidiFileEvent_getTrack(input_event)), 1), output_event_tick, MidiFileNoteOffEvent_getChannel(input_event), MidiFileNoteOffEvent_getNote(input_event), MidiFileNoteOffEvent_getVelocity(input_event));
		}
		case MIDI_FILE_EVENT_TYPE_NOTE_ON:
		{
			return MidiFileTrack_createNoteOnEvent(MidiFile_getTrackByNumber(output_midi_file, MidiFileTrack_getNumber(MidiFileEvent_getTrack(input_event)), 1), output_event_tick, MidiFileNoteOnEvent_getChannel(input_event), MidiFileNoteOnEvent_getNote(input_event), MidiFileNoteOnEvent_getVelocity(input_event));
		}
		case MIDI_FILE_EVENT_TYPE_KEY_PRESSURE:
		{
			return MidiFileTrack_createKeyPressureEvent(MidiFile_getTrackByNumber(output_midi_file, MidiFileTrack_getNumber(MidiFileEvent_getTrack(input_event)), 1), output_event_tick, MidiFileKeyPressureEvent_getChannel(input_event), MidiFileKeyPressureEvent_getNote(input_event), MidiFileKeyPressureEvent_getAmount(input_event));
		}
		case MIDI_FILE_EVENT_TYPE_CONTROL_CHANGE:
		{
			return MidiFileTrack_createControlChangeEvent(MidiFile_getTrackByNumber(output_midi_file, MidiFileTrack_getNumber(MidiFileEvent_getTrack(input_event)), 1), output_event_tick, MidiFileControlChangeEvent_getChannel(input_event), MidiFileControlChangeEvent_getNumber(input_event), MidiFileControlChangeEvent_getValue(input_event));
		}
		case MIDI_FILE_EVENT_TYPE_PROGRAM_CHANGE:
		{
			return MidiFileTrack_createProgramChangeEvent(MidiFile_getTrackByNumber(output_midi_file, MidiFileTrack_getNumber(MidiFileEvent_getTrack(input_event)), 1), output_event_tick, MidiFileProgramChangeEvent_getChannel(input_event), MidiFileProgramChangeEvent_getNumber(input_event));
		}
		case MIDI_FILE_EVENT_TYPE_CHANNEL_PRESSURE:
		{
			return MidiFileTrack_createChannelPressureEvent(MidiFile_getTrackByNumber(output_midi_file, MidiFileTrack_getNumber(MidiFileEvent_getTrack(input_event)), 1), output_event_tick, MidiFileChannelPressureEvent_getChannel(input_event), MidiFileChannelPressureEvent_getAmount(input_event));
		}
		case MIDI_FILE_EVENT_TYPE_PITCH_WHEEL:
		{
			return MidiFileTrack_createPitchWheelEvent(MidiFile_getTrackByNumber(output_midi_file, MidiFileTrack_getNumber(MidiFileEvent_getTrack(input_event)), 1), output_event_tick, MidiFilePitchWheelEvent_getChannel(input_event), MidiFilePitchWheelEvent_getValue(input_event));
		}
		case MIDI_FILE_EVENT_TYPE_SYSEX:
		{
			return MidiFileTrack_createSysexEvent(MidiFile_getTrackByNumber(output_midi_file, MidiFileTrack_getNumber(MidiFileEvent_getTrack(input_event)), 1), output_event_tick, MidiFileSysexEvent_getDataLength(input_event), MidiFileSysexEvent_getData(input_event));
		}
		case MIDI_FILE_EVENT_TYPE_META:
		{
			return MidiFileTrack_createMetaEvent(MidiFile_getTrackByNumber(output_midi_file, MidiFileTrack_getNumber(MidiFileEvent_getTrack(input_event)), 1), output_event_tick, MidiFileMetaEvent_getNumber(input_event), MidiFileMetaEvent_getDataLength(input_event), MidiFileMetaEvent_getData(input_event));
		}
		default:
		{
			return NULL;
		}
	}
}

static char * getoptarg(int argc, char **argv, int *indx)
{
	char *opt = argv[*indx];
	if (opt[1] == '-' || opt[2] == 0) { /* must be long arg or single char */
		if (++*indx == argc) usage(argv[0]);
		return argv[*indx];
	}
	opt += 2;	/* select value part of string */	
	return opt;
}

int main(int argc, char **argv)
{
	int i;
	int click_track_number = -1;
	int beats_per_click = 1;
	int beats_per_bar = 4;
	int replace_original = 0;
	char *output_filename = NULL;
	char *input_filename = NULL;
	MidiFile_t input_midi_file;
	MidiFile_t output_midi_file;
	float first_click_time;
	float first_bar_time=-1.0;
	float time_offset=0.0;
	float left_click_time;
	float right_click_time;
	MidiFileEvent_t click_event;
	float previous_output_tempo = -1.0; /* force at least one tempo event to be emitted */
	MidiFileEvent_t input_event;
	long first_voice_tick;

	/* The hard way, without 'getopt_long', so Windows might even be able to use it! */
	for (i = 1; i < argc; i++)
	{
		if (strcmp(argv[i], "--click-track") == 0 || strncmp(argv[i], "-c", 2) == 0)
		{
			click_track_number = atoi(getoptarg(argc, argv, &i));
		}
		else if (strcmp(argv[i], "--out") == 0 || strncmp(argv[i], "-o", 2) == 0)
		{
			output_filename = getoptarg(argc, argv, &i);
		}
		else if (strcmp(argv[i], "--beats-per-click") == 0 || strncmp(argv[i], "-b", 2) == 0)
		{
			beats_per_click = atoi(getoptarg(argc, argv, &i));
		}
		else if (strcmp(argv[i], "--beats-in-bar") == 0 || strncmp(argv[i], "-n", 2) == 0)
		{
			beats_per_bar = atoi(getoptarg(argc, argv, &i));
		}
		else if (strcmp(argv[i], "--replace") == 0 || strcmp(argv[i], "-r") == 0)
		{
			replace_original = 1;
		}
		else if (input_filename == NULL)
		{
			input_filename = argv[i];
		}
		else
		{
			usage(argv[0]);
		}
	}

	if ((click_track_number < 0) || (input_filename == NULL)) usage(argv[0]);
	if (output_filename == NULL) {
		if (replace_original)
			output_filename = input_filename;
		else {
			fprintf(stderr, "Error:  No output file or '--replace' option\n");
			usage(argv[0]);		
		}
	}

	if ((input_midi_file = MidiFile_load(input_filename)) == NULL)
	{
		fprintf(stderr, "Error:  Cannot read MIDI file \"%s\".\n", input_filename);
		exit(1);
	}

	if (MidiFile_getFileFormat(input_midi_file) != 1)
	{
		fprintf(stderr, "Error:  Because this algorithm makes use of multiple, simultaneous tracks, it requires MIDI files in format 1.\n");
		exit(1);
	}

	if (MidiFile_getDivisionType(input_midi_file) != MIDI_FILE_DIVISION_TYPE_PPQ)
	{
		fprintf(stderr, "Error:  Because this algorithm makes use of tempo events, it requires MIDI files which use the PPQ division type, not SMPTE.\n");
		exit(1);
	}

	output_midi_file = MidiFile_new(1, MIDI_FILE_DIVISION_TYPE_PPQ, MidiFile_getResolution(input_midi_file));

	click_event = get_first_click_event(input_midi_file, click_track_number);

	if (click_event == NULL)
	{
		fprintf(stderr, "Error:  No click found in track %d!\n", click_track_number);
		exit(1);
	}

	first_voice_tick = MidiFileEvent_getTick(click_event);
	first_click_time = left_click_time
	 = MidiFile_getTimeFromTick(input_midi_file, first_voice_tick);


	click_event = get_next_click_event(click_event);
	if (click_event == NULL)
	{
		fprintf(stderr, "Error:  Only one click found in track %d!\n", click_track_number);
		exit(1);
	}

	while (click_event)
	{
		float output_tempo;
		right_click_time = MidiFile_getTimeFromTick(input_midi_file, MidiFileEvent_getTick(click_event));
		output_tempo = (60.0 * beats_per_click) / (right_click_time - left_click_time);

			if (left_click_time == first_click_time) {
			long ticks_per_bar = beats_per_bar * MidiFile_getResolution(output_midi_file);
			long first_bar_tick = 0;
			MidiFileTrack_createTempoEvent(MidiFile_getTrackByNumber(output_midi_file, 0, 1),
				0, output_tempo);	/* Initial tempo */
			do {
				first_bar_time =
					MidiFile_getTimeFromTick(output_midi_file, first_bar_tick);
				first_bar_tick += ticks_per_bar;
			} while (first_bar_time < first_click_time);
			time_offset = first_bar_time - first_click_time;
			DPRINTF(("first bar %f, first click %f (resolution %d)\n",
				first_bar_time, first_click_time,
				MidiFile_getResolution(input_midi_file));)
		}

		else if (output_tempo != previous_output_tempo)
		{
			MidiFileTrack_createTempoEvent(MidiFile_getTrackByNumber(output_midi_file, 0, 1),
			 MidiFile_getTickFromTime(output_midi_file, left_click_time+time_offset), output_tempo);
			previous_output_tempo = output_tempo;
		}

		left_click_time = right_click_time;
		click_event = get_next_click_event(click_event);
	}

	for (input_event = MidiFile_getFirstEvent(input_midi_file); input_event != NULL; input_event = MidiFileEvent_getNextEventInFile(input_event))
	{
		long input_tick;
		if (MidiFileEvent_isTempoEvent(input_event)) continue;
		input_tick = MidiFileEvent_getTick(input_event);
		if (input_tick > first_voice_tick || MidiFileEvent_isNoteStartEvent(input_event) )
		{
			float event_time = MidiFile_getTimeFromTick(input_midi_file, input_tick);
			clone_event(input_event, output_midi_file, MidiFile_getTickFromTime(output_midi_file, event_time  + time_offset));
			if (input_tick < first_voice_tick) first_voice_tick = input_tick;
		}
		else {	/* leave pre-first-click non-channel events alone */
			clone_event(input_event, output_midi_file, input_tick);
		}
	}

	if (MidiFile_save(output_midi_file, output_filename) < 0)
	{
		fprintf(stderr, "Error:  Cannot write MIDI file \"%s\".\n", output_filename);
		exit(1);
	}

	return 0;
}

